from collections import namedtuple, defaultdict
import threading
import queue
from atexit import register

from kuai.backends import WeakCallback, Singleton, singleton_object


@singleton_object
class ThreadedBackend(metaclass=Singleton):
    handlers = defaultdict(set)
    events = queue.Queue()
    isRunning = False

    def startThread(self):
        self.__thread = threading.Thread(target=self.__engine)
        self.__thread.daemon = True
        self.__thread.start()
        register(self.__stop)

    def addHandler(self, event, callback, *args, **kwargs):
        if not self.isRunning:
            self.isRunning = True
            self.startThread()
        self.handlers[event].add(WeakCallback(callback))

    def handleEvent(self, event, *args, **kwargs):
        if not self.isRunning:
            self.isRunning = True
            self.startThread()
        Event = namedtuple("Event", "name args kwargs")
        self.events.put(Event(event, args, kwargs))

    def __stop(self):
        self.isRunning = False
        self.__thread.join()

    def __engine(self):
        while self.isRunning:
            try:
                currentEvent = self.events.get(timeout=0.3)
            except queue.Empty:
                pass
            else:
                callbacks = self.handlers.get(currentEvent.name)
                if callbacks is not None:
                    for callback in list(callbacks):
                        try:
                            callback(*currentEvent.args, **currentEvent.kwargs)
                        except Exception as e:
                            print("Kuai Exception in `threaded` backend : ", e)
                            raise
                self.events.task_done()


def setup(app):
    app.register_backend('threaded', ThreadedBackend)



from kuai import Kuai, set_backend


def test_set_backend():
    set_backend('simple')
    assert Kuai.which_backend == 'simple'
    set_backend('priority')
    assert Kuai.which_backend == 'priority'
    set_backend('simple')
    assert Kuai.which_backend == 'simple'



from collections import namedtuple, defaultdict

from kuai.backends import WeakCallback, Singleton, singleton_object


@singleton_object
class PriorityBackend(metaclass=Singleton):
    handlers = defaultdict(list)

    def addHandler(self, event, callback, priority=5):
        PriorityEvent = namedtuple("PriorityEvent", "priority callback")
        self.handlers[event].append(PriorityEvent(priority, WeakCallback(callback)))

    def handleEvent(self, event, *args, **kwargs):
        handlers = self.handlers.get(event)
        if handlers is None:
            return
        else:
            for handler in sorted(handlers, key=lambda t: t.priority):
                handler.callback(*args, **kwargs)


def setup(app):
    app.register_backend('priority', PriorityBackend)



# -*- coding: utf-8 -*-
"""Various helper functions for Kuai, these are all from stackoverflow. Amen."""
import weakref
import inspect


def singleton_object(cls):
    """Class decorator that transforms (and replaces) a class definition (which
    must have a Singleton metaclass) with the actual singleton object. Ensures
    that the resulting object can still be "instantiated" (i.e., called),
    returning the same object. Also ensures the object can be pickled, is
    hashable, and has the correct string representation (the name of the
    singleton)
    """
    assert isinstance(cls, Singleton), \
        cls.__name__ + " must use Singleton metaclass"

    def self_instantiate(self):
        return self

    cls.__call__ = self_instantiate
    cls.__hash__ = lambda self: hash(cls)
    cls.__repr__ = lambda self: cls.__name__
    cls.__reduce__ = lambda self: cls.__name__
    obj = cls()
    obj.__name__ = cls.__name__
    return obj


class Singleton(type):
    """Metaclass for singleton objects. Any instantiation of a
    Singleton class yields the exact same object, e.g.:

    >>> class MyClass(metaclass=Singleton):
            pass
    >>> a = MyClass()
    >>> b = MyClass()
    >>> a is b
    True
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = \
                super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def __instancecheck__(mcs, instance):
        if instance.__class__ is mcs:
            return True
        else:
            return isinstance(instance.__class__, mcs)


class WeakCallback(object):
    """A Weak Callback object that will keep a reference to
    the connecting object with weakref semantics.

    This allows object A to pass a callback method to object S,
    without object S keeping A alive. **Extended to work with
    unbound calls, ie: functions, lambdas..
    """
    def __init__(self, callback):
        """Create a new Weak Callback calling @callback"""
        if inspect.ismethod(callback):
            obj = callback.__self__
            attr = callback.__func__.__name__
        else:
            obj = callback
            attr = None
        self.__wref = weakref.ref(obj, self.__object_deleted)
        self.__callback_attr = attr

    def __call__(self, *args, **kwargs):
        obj = self.__wref()
        if obj:
            if self.__callback_attr:
                attr = getattr(obj, self.__callback_attr)
            else:
                attr = obj
            attr(*args, **kwargs)
        else:
            self.__default_callback(*args, **kwargs)

    def __default_callback(self, *args, **kwargs):
        """Called instead of callback when expired"""
        pass

    def __object_deleted(self, wref):
        """Called when callback expires"""
        pass



import collections
from functools import partial
from kuai.backends import WeakCallback, Singleton, singleton_object


@singleton_object
class SignalBackend(metaclass=Singleton):
    """Bind a signal to a variable. 

        Added v0.1.1 - Duroktar

        >>> from kuai import Kuai, set_backend
        >>> set_backend('signal')

        >>> def hello(name):
        >>>     print(f"Hello, {name}")

        >>> signal = Kuai.on('signal', hello)
        >>> signal("cymrow")
    
     """
    handlers = collections.defaultdict(list)

    def addHandler(self, event, callback, *args, **kw):
        self.handlers[event].append(WeakCallback(callback))
        return partial(self.handleEvent, event)

    def handleEvent(self, event, *args, **kwargs):
        for handler in self.handlers.get(event, []):
            handler(*args, **kwargs)


def setup(app):
    app.register_backend('signal', SignalBackend)



import os
from functools import partial

from pluginbase import PluginBase

from kuai.helpers import WeakCallback, Singleton, singleton_object

here = os.path.abspath(os.path.dirname(__file__))
get_path = partial(os.path.join, here)


plugin_base = PluginBase(package='kuai.plugins')
# ------------------------------------

__all__ = [
    'Manager', 'set_backend', 'which_backend',
    'WeakCallback', 'Singleton', 'singleton_object'
]


@singleton_object
class Manager(metaclass=Singleton):
    _backends = {}
    current_backend = None

    def __init__(self):
        self._source = plugin_base.make_plugin_source(
            searchpath=[here]
        )
        self.set_backend()

    @property
    def backend(self):
        return self._backend

    def list_backends(self):
        return self._source.list_plugins()

    def _load_backend(self, name):
        backend = self._source.load_plugin(name)
        backend.setup(self)

    def register_backend(self, name, backend):
        """A function a plugin can use to register a backend."""
        self._backends[name] = backend

    def set_backend(self, name='simple'):
        try:
            self._load_backend(name)
        except Exception as e:
            print("Kuai: {} is not a valid backend.. "
                  "Please select from,\n").format(name)
            for each in self._source.list_plugins():
                print(" - ", each)
            exit(1)
        else:
            self.current_backend = name
            self._backend = self._backends[name]


def set_backend(name):
    Manager.set_backend(name)


def which_backend():
    print("Backend:", Manager.backend.name)




from kuai import Kuai, set_backend

x = 0


def test_priority_backend():
    set_backend('priority')

    def increment():
        global x
        x += 1

    Kuai.on('test', increment)
    Kuai.emit('test')
    assert x == 1

    Kuai.emit('test')
    assert x == 2



from kuai import Kuai, set_backend
from time import sleep

x = 0


def test_kuai_events():
    set_backend('threaded')

    def increment():
        global x
        x += 1

    Kuai.on('test', increment)
    Kuai.emit('test')
    sleep(0.01)
    assert x == 1

    Kuai.emit('test')
    sleep(0.01)
    assert x == 2



import collections
from kuai.backends import WeakCallback, Singleton, singleton_object


@singleton_object
class SimpleBackend(metaclass=Singleton):
    handlers = collections.defaultdict(list)

    def addHandler(self, event, callback, *args, **kw):
        self.handlers[event].append(WeakCallback(callback))

    def handleEvent(self, event, *args, **kwargs):
        for handler in self.handlers.get(event, []):
            handler(*args, **kwargs)


def setup(app):
    app.register_backend('simple', SimpleBackend)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

with open('requirements_dev.txt') as test_requirements:
    test_requirements = test_requirements.readlines()


setup(
    name='Kuai',
    version='0.1.1',
    description="Simple & fast event library. MIT licensed",
    long_description=readme + '\n\n' + history,
    author="Scott Doucet",
    author_email='duroktar@gmail.com',
    url='https://github.com/Duroktar/kuai',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='kuai',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from kuai import Kuai, set_backend, kuai_backend
from kuai.backends import Manager


def test_backend():
    set_backend('simple')
    assert kuai_backend() == 'simple'


def test_kuai_singleton():
    a = Kuai()
    b = Kuai()
    c = Kuai()
    assert (a == b == c)


def test_backend_singleton():
    a = Manager()
    b = Manager()
    c = Manager()
    assert (a == b == c)



# -*- coding: utf-8 -*-
from kuai.helpers import Singleton, singleton_object
from kuai.backends import Manager, set_backend

__author__ = """Scott Doucet"""
__email__ = 'duroktar@gmail.com'
__version__ = '0.1.1'
__all__ = [
    'Kuai',
    'kuai_backend', 'set_backend'
]


@singleton_object
class Kuai(metaclass=Singleton):

    @classmethod
    def on(cls, event, callback, *args, **kwargs):
        return Manager.backend.addHandler(event, callback, *args, **kwargs)

    @classmethod
    def emit(cls, event, *args, **kwargs):
        Manager.backend.handleEvent(event, *args, **kwargs)

    @property
    def which_backend(self):
        return Manager.current_backend


class KuaiBase:
    def on(cls, event, callback, *args, **kwargs):
        Manager.backend.addHandler(event, callback, *args, **kwargs)

    def emit(cls, event, *args, **kwargs):
        Manager.backend.handleEvent(event, *args, **kwargs)


def kuai_backend():
    return str(Kuai.which_backend)


Kuài = Kuai
